/* m_invite.c: Invites the user to join a channel.
 * Copyright (C) 2005 by MusIRCd Development.
 * $Id: m_invite.c,v 1.39 2005/07/19 05:24:16 musirc Exp $
 */

#include "handlers.h"
#include "channel.h"
#include "channel_mode.h"
#include "client.h"
#include "hash.h"
#include "istring.h"
#include "ircd.h"
#include "numeric.h"
#include "send.h"
#include "config.h"
#include "server.h"
#include "modules.h"
#include "packet.h"

static void m_invite(struct Client *, struct Client *, int, char **);

struct Message invite_msgtab = {
  "INVITE", 3, MFLG_SLOW,
  {m_unregistered, m_invite, m_invite, m_invite}
};

#ifndef STATIC_MODULES
void
_modinit(void)
{
  mod_add_cmd(&invite_msgtab);
}

void
_moddeinit(void)
{
  mod_del_cmd(&invite_msgtab);
}

const char *_version = "$Revision: 1.39 $";
#endif

/* parv[0] - sender prefix
 * parv[1] - user to invite
 * parv[2] - channel name
 * parv[3] - invite timestamp
 */
static void
m_invite(struct Client *client_p, struct Client *source_p,
         int parc, char *parv[])
{
  struct Client *target_p = NULL;
  struct Channel *chptr = NULL;
  struct Membership *ms = NULL;

  if (IsServer(source_p))
    return;

  if (MyClient(source_p) && !IsFloodDone(source_p))
    flood_endgrace(source_p);

  if ((target_p = find_person(client_p, parv[1])) == NULL)
  {
    sendto_one(source_p, form_str(ERR_NOSUCHNICK),
               me.name, source_p->name, parv[1]);
    return;
  }

  /* Do not send local channel invites to users if they are not on the
   * same server as the person sending the INVITE message.
   * Possibly should be an error sent to source_p
   * done .. there should be no problem because MyConnect(source_p) should
   * always be true if parse() and such is working correctly --is
   */
  if (!MyConnect(target_p) && (*parv[2] == '&'))
  {
    if (ConfigServerHide.hide_servers == 0)
      sendto_one(source_p, form_str(ERR_USERNOTONSERV),
                 me.name, source_p->name, parv[1]);
    return;
  }

  if ((chptr = hash_find_channel(parv[2])) == NULL)
  {
    sendto_one(source_p, form_str(ERR_NOSUCHCHANNEL),
               me.name, source_p->name, parv[2]);
    return;
  }

  if ((ms = find_channel_link(source_p, chptr)) == NULL)
  {
    sendto_one(source_p, form_str(ERR_NOTONCHANNEL),
               me.name, source_p->name, parv[2]);
    return;
  }

  if (!has_member_flags(ms, CHFL_CHANOP|CHFL_HALFOP) &&
      (chptr->mode.mode & MODE_INVITEONLY))
  {
    sendto_one(source_p, form_str(ERR_CHANOPRIVSNEEDED),
               me.name, source_p->name, chptr->name);
    return;
  }

  if (IsMember(target_p, chptr))
  {
    sendto_one(source_p, form_str(ERR_USERONCHANNEL),
               me.name, source_p->name, parv[1], parv[2]);
    return;
  }

  if (MyConnect(source_p))
  {
    sendto_one(source_p, form_str(RPL_INVITING),
               me.name, source_p->name, target_p->name, parv[2]);

    if (target_p->away)
      sendto_one(source_p, form_str(RPL_AWAY),
                 me.name, source_p->name, target_p->name,
                 target_p->away);
  }
  else if (parc > 3 && IsDigit(*parv[3]))
    if (atoi(parv[3]) > chptr->channelts)
      return;

  if (MyConnect(target_p))
  {
    sendto_one(target_p, ":%s!%s@%s INVITE %s :%s",
               source_p->name, source_p->username,
               source_p->host, target_p->name, chptr->name);

    if (chptr->mode.mode & MODE_INVITEONLY)
    {
      if (chptr->mode.mode & MODE_PRIVATE)
      {
        /* Only do this if channel is set +i and +p */
        sendto_channel_local(CHFL_CHANOP|CHFL_HALFOP, chptr,
                             ":%s NOTICE %s :%s is inviting %s to %s.",
                             me.name, chptr->name, source_p->name,
                             target_p->name, chptr->name);
        sendto_channel_remote(source_p, CHFL_CHANOP|CHFL_HALFOP,
                              chptr, ":%s NOTICE %s :%s is inviting %s to %s.",
                              source_p->name, chptr->name, source_p->name,
                              target_p->name, chptr->name);
      }
      add_invite(chptr, target_p);
    }
  }
  else if (target_p->from != client_p)
    sendto_one(target_p, ":%s INVITE %s %s %lu",
               source_p->name, target_p->name,
               chptr->name, (unsigned long)chptr->channelts);
}
