/* m_whois.c: Shows who a user is.
 * Copyright (C) 2005 by MusIRCd Development.
 * $Id: m_whois.c,v 1.125 2005/07/19 05:24:18 musirc Exp $
 */
 
#include "whowas.h"
#include "handlers.h"
#include "hash.h"
#include "istring.h"
#include "sprintf.h"
#include "numeric.h"
#include "ircd.h"
#include "server.h"
#include "config.h"
#include "send.h"
#include "modules.h"
#include "channel.h"
#include "channel_mode.h"

static void do_whois(struct Client *, int, char *[], int);
static int single_whois(struct Client *, struct Client *);
static void whois_person(struct Client *, struct Client *, int);
static int global_whois(struct Client *, const char *);
static void m_whois(struct Client *, struct Client *, int, char *[]);
static void mo_whois(struct Client *, struct Client *, int, char *[]);
static void mo_owhois(struct Client *, struct Client *, int, char *[]);
static void m_whowas(struct Client *, struct Client *, int, char *[]);
static void mo_whowas(struct Client *, struct Client *, int, char *[]);
static void do_whowas(struct Client *, struct Client *, int, char *[]);

struct Message whois_msgtab = {
  "WHOIS", 0, MFLG_SLOW,
  {m_unregistered, m_whois, mo_whois, mo_whois}
};
struct Message owhois_msgtab = {
  "OWHOIS", 0, MFLG_SLOW,
  {m_ignore, m_ignore, m_ignore, mo_owhois}
};
struct Message whowas_msgtab = {
  "WHOWAS", 0, MFLG_SLOW,
  {m_unregistered, m_whowas, m_ignore, mo_whowas}
};

#ifndef STATIC_MODULES
void
_modinit(void)
{
  mod_add_cmd(&whois_msgtab);
  mod_add_cmd(&owhois_msgtab);
  mod_add_cmd(&whowas_msgtab);
}

void
_moddeinit(void)
{
  mod_del_cmd(&whois_msgtab);
  mod_del_cmd(&owhois_msgtab);
  mod_del_cmd(&whowas_msgtab);
}

const char *_version = "$Revision: 1.125 $";
#endif

/* parv[0] = sender prefix
 * parv[1] = nickname masklist
 */
static void
m_whois(struct Client *client_p, struct Client *source_p, int parc, char *parv[])
{
  static time_t last_used = 0;
  
  if (parc < 2 || EmptyString(parv[1]))
  {
    sendto_one(source_p, form_str(ERR_NONICKNAMEGIVEN),
               me.name, source_p->name);
    return;
  }

  /* seeing as this is going across servers, we should limit it */
  if ((last_used + ConfigFileEntry.pace_wait_simple) > CurrentTime)
  {
    sendto_one(source_p, form_str(RPL_LOAD2HI),
	       me.name, source_p->name);
    return;
  }
  else
    last_used = CurrentTime;

  /* if we have serverhide enabled, they can either ask the clients
   * server, or our server.. I dont see why they would need to ask
   * anything else for info about the client.. --fl_
   */
  if (parc > 2)
  {
    if (hunt_server(client_p,source_p,":%s WHOIS %s :%s", 1, parc, parv) != HUNTED_ISME)
      return;

    parv[1] = parv[2];
  }
  do_whois(source_p, parc, parv, 0);
}

/* parv[0] = sender prefix
 * parv[1] = nickname masklist
 */
static void
mo_whois(struct Client *client_p, struct Client *source_p,
         int parc, char *parv[])
{
  if (parc < 2 || EmptyString(parv[1]))
  {
    sendto_one(source_p, form_str(ERR_NONICKNAMEGIVEN),
               me.name, source_p->name);
    return;
  }
  if (parc > 2)
  {
    if (hunt_server(client_p,source_p,":%s WHOIS %s :%s", 1, parc, parv) != HUNTED_ISME)
      return;

    parv[1] = parv[2];
  }
  do_whois(source_p, parc, parv, 0);
}

/* parv[0] = sender prefix
 * parv[1] = nickname masklist
 */
static void
mo_owhois(struct Client *client_p, struct Client *source_p, int parc, char *parv[])
{
  struct Client *target_p;

  if (parc < 2)
  {
    sendto_one(source_p, form_str(ERR_NONICKNAMEGIVEN),
               me.name, source_p->name);
    return;
  }
  do_whois(source_p, parc, parv, 1);

  if ((target_p = find_client(parv[1])) != NULL)
    sendto_realops_flags(UMODE_SPY, L_ALL, "OWHOIS '%s' by %s (%s@%s) [%s]",
                         target_p->name, source_p->name, source_p->username,
                         source_p->host, source_p->servptr->name);
}

static void
do_whois(struct Client *source_p, int parc, char *parv[], int owhois)
{
  struct Client *target_p;
  char *nick, *p = NULL;
  int found = 0;

  nick = parv[1];
  while (*nick == ',')
    nick++;
  if ((p = strchr(nick,',')) != NULL)
    *p = '\0';

  if (*nick == '\0')
    return;

  collapse(nick);

  if (strpbrk(nick, "?#*") == NULL)
  {
    if ((target_p = find_client(nick)) != NULL)
    {
      if (IsClient(target_p))
      {
        whois_person(source_p, target_p, owhois);
        found = 1;
      }
    }
  }
  else /* wilds is true */
  {
    if (MyClient(source_p))
      found = global_whois(source_p, nick);
  }

  if (!found)
    sendto_one(source_p, form_str(ERR_NOSUCHNICK),
	       me.name, source_p->name, parv[1]);
  sendto_one(source_p, form_str(RPL_ENDOFWHOIS),
	     me.name, source_p->name, parv[1]);
}

/* Inputs	- source_p client to report to
 *		- target_p client to report on
 * Output	- if found return 1
 * Side Effects	- do a single whois on given client
 * 		  writing results to source_p
 */
static int
global_whois(struct Client *source_p, const char *nick)
{
  dlink_node *ptr;
  struct Client *target_p;
  int found = 0;

  DLINK_FOREACH(ptr, global_client_list.head)
  {
    target_p = ptr->data;

    if (!IsClient(target_p))
      continue;
 
    if (!match(nick, target_p->name))
      continue;

    /* 'Rules' established for sending a WHOIS reply:
     * - if wildcards are being used dont send a reply if
     *   the querier isnt any common channels and the
     *   client in question is invisible and wildcards are
     *   in use (allow exact matches only);
     *
     * - only send replies about common or public channels
     *   the target user(s) are on;
     */
    found |= single_whois(source_p, target_p);
  }
  return(found);
}

/* Inputs       - source_p client to report to
 *              - target_p client to report on
 * Output       - if found return 1
 * Side Effects - do a single whois on given client
 *                writing results to source_p
 */
static int
single_whois(struct Client *source_p, struct Client *target_p)
{
  dlink_node *ptr;
  struct Channel *chptr;

  if (!IsInvisible(target_p) || target_p == source_p)
  {
    /* always show user if they are visible (no +i) */
    whois_person(source_p, target_p, 0);
    return 1;
  }

  /* target_p is +i. Check if it is on any common channels with source_p */
  DLINK_FOREACH(ptr, target_p->channel.head)
  {
    chptr = ((struct Membership *)ptr->data)->chptr;
    if (IsMember(source_p, chptr)) 
    {
      whois_person(source_p, target_p, 0);
      return 1;
    }
  }
  return 0;
}

/* inputs       - source_p client to report to
 *              - target_p client to report on
 */
static void
whois_person(struct Client *source_p, struct Client *target_p, int owhois)
{
  char buf[BUFSIZE], *t = NULL;
  dlink_node *lp;
  struct Client *server_p;
  struct Channel *chptr;
  struct Membership *ms;
  int cur_len = 0, mlen, tlen, reply_to_send = 0;

  server_p = target_p->servptr;

  sendto_one(source_p, form_str(RPL_WHOISUSER), me.name,
         source_p->name, target_p->name,
         target_p->username, target_p->host, target_p->info);

  if (!IsService(target_p))
  {
    cur_len = mlen = ircsprintf(buf, form_str(RPL_WHOISCHANNELS),
		                me.name, source_p->name, target_p->name, "");
    t = buf + mlen;

    DLINK_FOREACH(lp, target_p->channel.head)
    {
      ms = lp->data;
      chptr = ms->chptr;

      if (ShowChannel(source_p, chptr) && (owhois == 0))
      {
        if ((cur_len + 3 + strlen(chptr->name) + 1) > (BUFSIZE - 2))
        {
          *(t - 1) = '\0';
          sendto_one(source_p, "%s", buf);
          cur_len = mlen;
          t = buf + mlen;
        }
        tlen = ircsprintf(t, "%s%s ", get_member_status(ms, 1), chptr->name);
        t += tlen;
        cur_len += tlen;
        reply_to_send = 1;
      }
      else if (owhois)
      {
        if ((cur_len + 3 + strlen(chptr->name) + 1) > (BUFSIZE - 2))
        {
          *(t - 1) = '\0';
          sendto_one(source_p, "%s", buf);
          cur_len = mlen;
          t = buf + mlen;
        }
        tlen = ircsprintf(t, "%s%s%s ", ShowChannel(source_p, chptr) ? "" : "^",
                          get_member_status(ms, 1), chptr->name);
        t += tlen;
        cur_len += tlen;
        reply_to_send = 1;
      }
    }
    if (reply_to_send)
    {
      *(t - 1) = '\0';
      sendto_one(source_p, "%s", buf);
    }
  }

  if (IsOper(source_p) || !ConfigServerHide.hide_servers || target_p == source_p)
    sendto_one(source_p, form_str(RPL_WHOISSERVER),
               me.name, source_p->name, target_p->name, server_p->name,
	       server_p->info);
  else
    sendto_one(source_p, form_str(RPL_WHOISSERVER),
               me.name, source_p->name, target_p->name,
               ServerInfo.network_name, server_p->info);

  if (target_p->away != NULL)
    sendto_one(source_p, form_str(RPL_AWAY), me.name,
               source_p->name, target_p->name, target_p->away);

  if (IsOper(target_p))
    sendto_one(source_p, form_str(IsAdmin(target_p) ? RPL_WHOISADMIN :
               RPL_WHOISOPERATOR), me.name, source_p->name, target_p->name);

  if (IsService(target_p))
  {
    sendto_one(source_p, form_str(RPL_WHOISSERVICE),
               me.name, source_p->name, target_p->name);
    return;
  }

  if (MyConnect(target_p))
  {
    if (IsOper(target_p) && (target_p != source_p) && target_p->umodes & UMODE_SPY)
       sendto_one(target_p, ":%s NOTICE %s :WHOIS by %s (%s@%s) [%s]",
                  me.name, source_p->name, source_p->name, source_p->username,
                  source_p->host, source_p->servptr->name);

    if (!IsIPSpoof(target_p) &&
	!(ConfigFileEntry.hide_spoof_ips && IsIPSpoof(target_p)) &&
        (target_p->sockhost[0] != '\0') && MyConnect(target_p))
    {
      if (IsAdmin(source_p) || source_p == target_p)
        sendto_one(source_p, form_str(RPL_WHOISACTUALLY),
                   me.name, source_p->name, target_p->name,
		   target_p->sockhost);
      else
        sendto_one(source_p, form_str(RPL_WHOISACTUALLY),
                   me.name, source_p->name, target_p->name,
                   IsIPSpoof(target_p) ? "masked" : target_p->sockhost);
    }
    sendto_one(source_p, form_str(RPL_WHOISIDLE),
               me.name, source_p->name, target_p->name,
               CurrentTime - target_p->localClient->last,
               target_p->firsttime);
  }
}

/* parv[0] = sender prefix
 * parv[1] = nickname queried
 */
static void
m_whowas(struct Client *client_p, struct Client *source_p,
         int parc, char *parv[])
{
  static time_t last_used = 0;

  if (parc < 2 || EmptyString(parv[1]))
  {
    sendto_one(source_p, form_str(ERR_NONICKNAMEGIVEN),
               me.name, source_p->name);
    return;
  }

  if((last_used + ConfigFileEntry.pace_wait_simple) > CurrentTime)
  {
    sendto_one(source_p,form_str(RPL_LOAD2HI),
 	       me.name,source_p->name);
    return;
  }
  else
    last_used = CurrentTime;
  do_whowas(client_p,source_p,parc,parv);
}

static void
mo_whowas(struct Client *client_p, struct Client *source_p,
          int parc, char *parv[])
{
  if (parc < 2 || EmptyString(parv[1]))
  {
    sendto_one(source_p, form_str(ERR_NONICKNAMEGIVEN),
               me.name, source_p->name);
    return;
  }
  do_whowas(client_p,source_p,parc,parv);
}

static void
do_whowas(struct Client *client_p, struct Client *source_p,
          int parc, char *parv[])
{
  struct Whowas *temp;
  int cur = 0, max = -1, found = 0;
  char *p, *nick;

  if (parc > 2)
    max = atoi(parv[2]);
  if (parc > 3)
    if (hunt_server(client_p,source_p,":%s WHOWAS %s %s :%s", 3,parc,parv))
      return;

  nick = parv[1];
  while (*nick == ',')
    nick++;
  if((p = strchr(nick,',')) != NULL)
    *p = '\0';
  if (!*nick)
    return;

  temp = WHOWASHASH[hash_whowas_name(nick)];

  for(;temp;temp=temp->next)
  {
    if (!irccmp(nick, temp->name))
    {
      sendto_one(source_p, form_str(RPL_WHOWASUSER),
                 me.name, source_p->name, temp->name,
                 temp->username, temp->hostname, temp->realname);

       if (ConfigServerHide.hide_servers && !IsOper(source_p))
         sendto_one(source_p, form_str(RPL_WHOISSERVER),
                    me.name, source_p->name, temp->name,
                    ServerInfo.network_name, myctime(temp->logoff));
       else
	 sendto_one(source_p, form_str(RPL_WHOISSERVER),
                    me.name, source_p->name, temp->name,
                    temp->servername, myctime(temp->logoff));
       cur++;
       found++;
     }
     if (max > 0 && cur >= max)
       break;
  }
  if (!found)
    sendto_one(source_p, form_str(ERR_WASNOSUCHNICK),
               me.name, source_p->name, parv[1]);

  sendto_one(source_p, form_str(RPL_ENDOFWHOWAS),
	     me.name, source_p->name, parv[1]);
}
